#!/usr/bin/perl

#
# Linux iSCSI Enterprise Target Portal (iportal)
#
# Steffen Plotner, 2010
#
# Released 2010.07.28
#
# This portal queries multiple iet instances to determine which targets
# the initiator should see (iet itself solves that also). However, we
# now have to in iet define which ports the target is reachable by. This
# causes issues when initiators like open-iscsi connect multiple times
# if a target is exposed on 2 IP addresses. The behavior is correct however
# it forces multipathing on the initiator host configuration which is not
# always desired. So, this portal recognizes that and only hands out one
# IP that is valid and functional to the initiators at a time. Initiators that have
# actually multiple source IP address like vmware will get one session per
# IP address (and not 2*2=4 sessions as we would be without the portal).
#
# iportal has 2 major code paths:
#
# 1) It is an iscsi initiator to iet's iscsi target port to obtain the 
#    listing of all TargetNames and their TargetAddresses.
# 2) It is an iscsi target portal to frontend iet's iscsi target in 
#    order to handle initiator logon and SendTargets=All command - nothing else.
#
# The iportal has been modified to support iscsi redirects, hence on initial
# login and sendtargets=all the initiator will receive targets that are
# located at the portal. Once the initiator attempts to connect to those targets
# the initiator will be given a redirect to the actual target - this allows us
# to implement smoother failover from node1 to node2, since we can simply
# stop iet, which causes an async logout request sent from the target to the
# initiator, which in turn will attempt to logon to the portal. The portal will
# respond with no TargetAddress values, hence the initiator keeps retrying
# until the portal has valid TargetAddresses again (i.e. when it failed over
# and the underlying resources are back online).
#
# Shortcomings:
#
# * iSNS would also have worked, except that vmware esx does not support iSNS.
# * iportal does not support chap, however the target connection can use chap
#
# We had attempted to put a RedirectExec feature into ietd, however, that turned
# into a disaster when you have 50+ initiators attempting to re-connect after
# target resources moved, ietd is single threaded by design - oops.
#
# Initiators tested:
#
# CentOS5, CentOS4, RedHat AS3, Windows 2003, Windows 2008, VMware ESX 3, VMware ESXi 4
# Atto for Mac, QLA4010
#
use strict;
use Data::Dumper;
use Getopt::Std;
use IO::Socket;
use POSIX qw(setsid SIGTERM SIGKILL :sys_wait_h);
use Cwd qw(abs_path);
use Fcntl qw(:flock);
use Sys::Syslog;

use FindBin;
use lib "$FindBin::Bin/lib";
use lib "$FindBin::Bin/etc";

use config;
use iscsi;
use iscsiportal;
use logutil;

#
# filenames that we use
#
my $pid_file = abs_path("/var/run/iscsi_portal_daemon.pid");

#
# iet's default is 1, so we have to use that also
#
my $target_portal_group_tag = 1;

#
# children processes that we own (just track key=pid and no value)
#
my %children_pids;
my $is_working = 1;

#
# handle command line options
#
my %opts;
getopts('hvdbfkl:',\%opts);
if ( defined($opts{'h'}) )
{
        print <<EOH;
$0 [-h] [-v] [-d] -i iqn | -f | -b | -k | -l

-h	help

-b	run the iscsi portal in the background (daemon)
-f	run the iscsi portal in the foreground
-k	kill background daemon

-v	verbose
-d	debug PDUs

-l	listen on port number other than the default ($cfg{'listen_port'})
EOH
	exit(0);
}

if ( defined($opts{'v'}) )
{
	$cfg{'verbose'} = 1;
}
$cfg{'debug'} = 0;
if ( defined($opts{'d'}) )
{
	$cfg{'debug'} = 1;
}

if ( defined($opts{'l'}) )
{
	$cfg{'listen_port'} = $opts{'l'};
}

if ( defined($opts{'b'}) )
{
	daemonize();
	iportal_listen();
	exit(0);
}
elsif ( defined($opts{'k'}) )
{
	iportal_daemon_kill();
}
elsif ( defined($opts{'f'}) )
{
	iportal_listen();
	exit(0);
}
else
{
	print "-h for more info\n";
}
exit(0);


#------------------------------------------------------------------------------
# daemonize
#------------------------------------------------------------------------------
sub daemonize
{
	chdir '/' or die "Can't chdir to /: $!";
	open STDIN, '/dev/null'	or die "Can't read /dev/null: $!";
	open STDOUT, '>/dev/null' or die "Can't write to /dev/null: $!";
	open STDERR, '>/dev/null' or die "Can't write to /dev/null: $!";
	defined(my $pid = fork) or die "Can't fork: $!";
	exit if $pid;
	setsid or die "Can't start a new session: $!";
	umask 0;

	#
	# re-open stdout and stderr
	#
	#open STDOUT, ">>$log_stdout";
	#open STDERR, ">>$log_stderr";
	
	#
	# register the pid
	#
	iportal_daemon_pid();
}

#------------------------------------------------------------------------------
# sig child handler
#------------------------------------------------------------------------------
sub sig_child
{
	while (( my $pid = waitpid(-1, WNOHANG)) != -1)
	{
		#
		# remove the pid from the children_pids
		#
		if ( $pid )
		{
			logger("sig_child> pid: $pid gone");
			delete $children_pids{$pid};
		}
		else
		{
			#
			# pid was zero, do nothing
			#
			last;
		}
	}
	$SIG{'CHLD'} = \&sig_child;
}

#------------------------------------------------------------------------------
# sig term handler
#------------------------------------------------------------------------------
sub sig_term
{
	logger("sig_term> ending loop...");
	$is_working = 0;
}

#------------------------------------------------------------------------------
# iscsi portal target server
#------------------------------------------------------------------------------
sub iportal_listen
{
	#
	# setup signals
	#
	$SIG{'CHLD'} = \&sig_child;
	$SIG{'TERM'} = \&sig_term;
	$SIG{'INT'} = \&sig_term;
	$SIG{'PIPE'} = 'IGNORE';

	#
	# listen on socket
	#
	my $server_port = $cfg{'listen_port'};
	my $so_max_conn = SOMAXCONN;
	logger_info("portald> listening on port: $server_port so_max_conn: ".$so_max_conn);
	my $server = IO::Socket::INET->new(
					LocalAddr	=> '0.0.0.0',
					LocalPort	=> $server_port,
					Type		=> SOCK_STREAM,
					ReuseAddr	=> 1,
					Listen		=> $so_max_conn,
				);
	if ( ! $server )
	{
		logger_err("portald_err> unable to listen on socket with $server_port"); 
		exit(1);
	}
	
	#
	# accept connections and fork
	#
	logger("portald> parent pid: $$");
	while ($is_working)
	{
		#
		# due to the CHLD signal above, the accept() call might terminate
		# with a null client (http://www.perlmonks.org/?node_id=802731)
		#
		my $client = $server->accept();
		if ( !$client )
		{
			logger("portald> accept got interrupted, ready");
			next;
		}
		
		my $fpid = fork();
		if ( !defined($fpid) )
		{
			logger_err("portald_err> unable to fork");
			next;
		}
		elsif ( $fpid == 0 )
		{
			my $pid = $$;
			logger("portald> child process pid: $pid starting");
			iportal_session($client);
			logger("portald> child process pid: $pid complete");
			exit(0);
		}
		else
		{
			close($client);
			logger("portald> parent has launched pid: $fpid (track child pid)");
			$children_pids{$fpid} = undef;
		}
	}
	logger("portald> worker loop has terminated");
	
	#
	# close listening socket
	#
	logger("portald> closing listening socket");
	close($server);

	#
	# kill all remaining children
	#
	foreach my $pid (keys %children_pids)
	{
		logger("portald> killing child pid: $pid");
		kill SIGKILL, $pid;
	}
	
	logger("portald> closing log, bye");
	closelog();
}

#------------------------------------------------------------------------------
# manage the session
#------------------------------------------------------------------------------
sub iportal_session
{
	my ($client) = @_;

	my $s_ip = $client->peerhost();

	if ( !iportal_dowork($client) )
	{
		logger_err("portald_err> client disappeared: $s_ip");
	}

	iscsi_net_disconnect($client);
	logger("portald> initiator from ip: $s_ip closed");
	logger_access($s_ip, "disconnected");
}

#------------------------------------------------------------------------------
# iscsi process login (no auth) and sendtargets=All request from initiator
#------------------------------------------------------------------------------
sub iportal_dowork
{
	my ($socket) = @_;

	my $s_ip = $socket->peerhost();
	my $d_ip = $socket->sockhost();
	logger("portald> initiator from ip: $s_ip connected with: $d_ip");
	logger_access($s_ip, "connect to: $d_ip");

	my %pdu;
	my $packet;

	#
	# iscsi LOGIN
	#
	logger("portald> LOGIN");
	$pdu{'cmd'} = $iscsi_def{'CMD_RECV'};
	$packet = iscsi_net_recv($socket);
	if ( !defined($packet) )
	{
		logger_err("portald_err> invalid packet");
		return undef;
	}

	iscsi_pdu_login(\%pdu, $packet);
	if ( $pdu{'opcode'} != $iscsi_def{'OPCODE_LOGIN'} )
	{
		logger_err("portald_err> expecting Login got opcode: $pdu{'opcode'}");
		return undef;
	}

	#
	# from the keyvalue array pull out the initiator name, alias and session
	# type, the rest gets returned to the client.
	#
	my $initiator_name;
	my $session_type;
	my $target_desired;
	my @keyvals_out;
	foreach my $keyval (@{$pdu{'keyvals'}})
	{
		# items listed here are removed and noted
		if ( $keyval =~ /^InitiatorName=(\S+)$/ )
		{
			$initiator_name = $1;
			logger("portald> $keyval (matched)");
		}
		elsif ( $keyval =~ /^SessionType=(\S+)$/ )
		{
			$session_type = $1;
			logger("portald> $keyval (matched)");
		}
		elsif ( $keyval =~ /^TargetName=(\S+)$/ )
		{
			$target_desired = $1;
			logger("portald> $keyval (matched - going to redirect)");
		}
		elsif ( $keyval =~ /^AuthMethod=.*$/ )
		{
			logger("portald> $keyval (matched)");
		}
		elsif ( 
			# items listed here are silently removed
			! (
				$keyval =~ /^InitiatorAlias=\S+$/
			))
		{
			logger("portal_debug> $keyval (keep)");
			push @keyvals_out, $keyval;
		}
	}
	$pdu{'keyvals'} = \@keyvals_out;

	logger_access($s_ip, "initiator: $initiator_name (session: $session_type)");

	if ( !defined($initiator_name) )
	{
		logger_err("portald_err> expecting InitiatorName in keyvals section");
		return undef;
	}
	
	#logger_access($s_ip, "session type: $session_type");

	#
	# if a target name was specified in this logon event, tell the initiator
	# to redirect to the actual TargetAddress(es)
	#
	if ( defined($target_desired) )
	{
		#
		# TARGET CONNECTION
		#
		logger("portald> target_desired: $target_desired");
		logger_access($s_ip, "target: $target_desired");

		#
		# check if the requested target is on the targets_removed list and refuse it
		#
		my @resultant_remove_list = grep(/^$target_desired$/, 
			@{$cfg{'targets_removed'}});
		if ( scalar(@resultant_remove_list) == 1 )
		{
			logger_access($s_ip, "target found on target_removed list");
			$pdu{'keyvals'} = undef;

			#
			# iscsi LOGIN RESPONSE (target removed)
			#
			logger("portald> LOGIN RESPONSE");
			$pdu{'cmd'} = $iscsi_def{'CMD_SEND'};
			$pdu{'opcode'} = $iscsi_def{'OPCODE_LOGIN_RESPONSE'};
			$pdu{'opcode_flag'} = 0x00;
			$pdu{'status'} = 0x0204;
			iscsi_pdu_target_sn(\%pdu);
			iscsi_pdu_login_response(\%pdu, \$packet);
			if (!iscsi_net_send($socket, \$packet))
			{
				logger_err("portald_err> unable to send to network?");
				return undef;
			}

			return 1;
		}

		#
		# based on the source/dest ip, the initiator and the desired target
		# produce a target portal value (ip:port)
		#
		$pdu{'keyvals'} = undef;
		my $redirect_normal = 1;
		my $target_portal = initiator_target_portal($s_ip, $d_ip, $initiator_name, $target_desired);
		if ( ! defined($target_portal) )
		{
			#
			# if there is no target portal down, return the iscsi portal instead (loop!)
			#
			my $server_ip = $socket->sockhost();
			my $server_port = $socket->sockport();

			$target_portal = "$server_ip:$server_port";
			$redirect_normal = 0;
		}

		#
		# on a iscsi redirect, do not return the portal tag (spec)
		#
		push @{$pdu{'keyvals'}}, "TargetAddress=$target_portal";

		#
		# delay the response by 2 seconds - reason, when you offline stmf,
		# all initiators will hound you in this code path nonstop - so - pause
		#
		if ( ! $redirect_normal )
		{
			logger("portald> pausing redirect response by 2 seconds");
			select(undef,undef,undef,2);
		}
		
		#
		# iscsi LOGIN RESPONSE (moved temporarily)
		#
		logger("portald> LOGIN RESPONSE");
		$pdu{'cmd'} = $iscsi_def{'CMD_SEND'};
		$pdu{'opcode'} = $iscsi_def{'OPCODE_LOGIN_RESPONSE'};
		$pdu{'opcode_flag'} = 0x00;
		$pdu{'status'} = 0x0101;
		iscsi_pdu_target_sn(\%pdu);
		iscsi_pdu_login_response(\%pdu, \$packet);
		if (!iscsi_net_send($socket, \$packet))
		{
			logger_err("portald_err> unable to send to network?");
			return undef;
		}

		return 1;
	}
	
	#
	# determine the current and next stage of logon negotiation
	#
	# next stage indicates how we respond and what we are listening
	# for. The next stage is either login operational or full phase
	# 
	my $csg = $pdu{'opcode_flag'} & $iscsi_def{'ISCSI_CSG_MASK'};
	my $nsg = $pdu{'opcode_flag'} & $iscsi_def{'ISCSI_NSG_MASK'};
	logger("portald> csg: $csg nsg: $nsg");

	logger("portald> nsg: SECURITY_NEGOTIATION") if ( $nsg == $iscsi_def{'ISCSI_NSG_SECURITY_NEGOTIATION'} );
	logger("portald> nsg: LOGINOPERATIONAL_NEGOTIATION") if ( $nsg == $iscsi_def{'ISCSI_NSG_LOGINOPERATIONAL_NEGOTIATION'} );
	logger("portald> nsg: FULLFEATUREPHASE") if ( $nsg == $iscsi_def{'ISCSI_NSG_FULLFEATUREPHASE'} );

	#
	# looks like qla4010, when configured with CHAP forces us to use SECURITY_NEGOTIATION
	# for which we have no code - so those will not do chap. They come in with Auth=CHAP,None
	# This only appears to be the case for the boot device - however. The SANSurfer tool
	# later distinguished the discovery from the target session in terms of CHAP
	#

	#
	# initiator wants to negotiate logon
	#
	if ( $nsg == $iscsi_def{'ISCSI_NSG_LOGINOPERATIONAL_NEGOTIATION'} )
	{
		#
		# iscsi LOGIN RESPONSE
		#
		logger("portald> LOGIN RESPONSE");
		$pdu{'cmd'} = $iscsi_def{'CMD_SEND'};
		$pdu{'opcode'} = $iscsi_def{'OPCODE_LOGIN_RESPONSE'};
		$pdu{'opcode_flag'} = 
			$iscsi_def{'ISCSI_TRANSIT_ON'} |
			$iscsi_def{'ISCSI_CONTINUE_OFF'} |
			$csg |
			$nsg;
		$pdu{'status'} = 0;
		push @{$pdu{'keyvals'}}, 'AuthMethod=None';
		iscsi_pdu_target_sn(\%pdu);
		iscsi_pdu_login_response(\%pdu, \$packet);
		if (!iscsi_net_send($socket, \$packet))
		{
			logger_err("portald_err> unable to send to network?");
			return undef;
		}

		#
		# iscsi LOGIN
		#
		logger("portald> LOGIN");
		$pdu{'cmd'} = $iscsi_def{'CMD_RECV'};
		$packet = iscsi_net_recv($socket);
		if ( !defined($packet) )
		{
			logger_err("portald_err> invalid packet");
			return undef;
		}

		iscsi_pdu_login(\%pdu, $packet);
		if ( $pdu{'opcode'} != $iscsi_def{'OPCODE_LOGIN'} )
		{
			logger_err("portald_err> expecting Login got opcode: $pdu{'opcode'}");
			return undef;
		}

		$csg = $pdu{'opcode_flag'} & $iscsi_def{'ISCSI_CSG_MASK'};
		$nsg = $pdu{'opcode_flag'} & $iscsi_def{'ISCSI_NSG_MASK'};
		logger("portald> csg: $csg nsg: $nsg");

		logger("portald> nsg: SECURITY_NEGOTIATION") if ( $nsg == $iscsi_def{'ISCSI_NSG_SECURITY_NEGOTIATION'} );
		logger("portald> nsg: LOGINOPERATIONAL_NEGOTIATION") if ( $nsg == $iscsi_def{'ISCSI_NSG_LOGINOPERATIONAL_NEGOTIATION'} );
		logger("portald> nsg: FULLFEATUREPHASE") if ( $nsg == $iscsi_def{'ISCSI_NSG_FULLFEATUREPHASE'} );
	}

	#
	# iscsi LOGIN RESPONSE
	#
	logger("portald> LOGIN RESPONSE");
	$pdu{'cmd'} = $iscsi_def{'CMD_SEND'};
	$pdu{'opcode'} = $iscsi_def{'OPCODE_LOGIN_RESPONSE'};
	$pdu{'opcode_flag'} = 
		$iscsi_def{'ISCSI_TRANSIT_ON'} |
		$iscsi_def{'ISCSI_CONTINUE_OFF'} |
		$csg |
		$nsg;
	$pdu{'status'} = 0;
	iscsi_pdu_target_sn(\%pdu,0);
	iscsi_pdu_login_response(\%pdu, \$packet);
	if (!iscsi_net_send($socket, \$packet))
	{
		logger_err("portald_err> unable to send to network?");
		return undef;
	}

	#
	# iscsi TEXT COMMAND/LOGOUT
	#
	logger("portald> TEXT COMMAND");
	$pdu{'cmd'} = $iscsi_def{'CMD_RECV'};
	$packet = iscsi_net_recv($socket);
	if ( !defined($packet) )
	{
		logger_err("portald_err> invalid packet");
		return undef;
	}

	iscsi_pdu_text_command(\%pdu, $packet);	
	if ( $pdu{'opcode'} != $iscsi_def{'OPCODE_TEXT_COMMAND'} )
	{
		logger_err("portald_err> expecting Text Command got opcode: $pdu{'opcode'}");
		return undef;
	}

	#
	# check for SendTargets=All
	#
	my $found = 0;
	foreach my $keyval (@{$pdu{'keyvals'}})
	{
		if ( $keyval =~ /SendTargets=All/ )
		{
			logger_access($s_ip, "initiator request: $keyval");
			$found = 1;
		}
	}
	if ( !$found )
	{
		logger_err("portald_err> expecting SendTargets=All");
		return undef;
	}

	#
	# based on the initiator name, determine the related targets
	#
	my $discovered_targets = initiator_discover_targets($d_ip, $initiator_name);

	#
	# via iscsi we return the discovery response to the initiator
	# changing the target address to the portal for each target
	#
	$pdu{'keyvals'} = undef;
	foreach my $target (keys %{$discovered_targets} )
	{
		#
		# check if the requested target is on the targets_removed list and refuse it
		#
		my @resultant_remove_list = grep(/^$target$/, 
			@{$cfg{'targets_removed'}});
		if ( scalar(@resultant_remove_list) == 1 )
		{
			logger("portald> skipping $target");
		}
		else
		{
			my $val = "TargetName=$target";
			logger("portald> $val");
			push @{$pdu{'keyvals'}}, $val;

			my $server_ip = $socket->sockhost();
			my $server_port = $socket->sockport();
			$val = "TargetAddress=$server_ip:$server_port,$target_portal_group_tag";
			logger("portald> $val");
			push @{$pdu{'keyvals'}}, $val;
		}
	}

	#
	# log the found targets
	#
	foreach my $keyval (@{$pdu{'keyvals'}})
	{
		logger_access($s_ip, "portal response: $keyval");
	}

	#
	# iscsi TEXT RESPONSE
	#
	$pdu{'cmd'} = $iscsi_def{'CMD_SEND'};
	$pdu{'opcode'} = $iscsi_def{'OPCODE_TEXT_RESPONSE'};
	iscsi_pdu_target_sn(\%pdu,1);
	iscsi_pdu_text_response(\%pdu, \$packet);
	if (!iscsi_net_send($socket, \$packet))
	{
		logger_err("portald_err> unable to send to network?");
		return undef;
	}

	#
	# iscsi LOGOUT REQUEST
	#
	logger("portald> LOGOUT REQUEST");
	$pdu{'cmd'} = $iscsi_def{'CMD_RECV'};
	$packet = iscsi_net_recv($socket);
	if ( !defined($packet) )
	{
		logger("portald> initiator did not send logout request, closing connection");
		return 1;
	}

	iscsi_pdu_logout_request(\%pdu, $packet);
	if ( $pdu{'opcode'} == $iscsi_def{'OPCODE_LOGOUT_REQUEST'} )
	{
		#
		# iscsi LOGOUT RESPONSE
		#
		logger("portald> initiator is requesting to logout... bye");
		$pdu{'cmd'} = $iscsi_def{'CMD_SEND'};
		$pdu{'opcode'} = $iscsi_def{'OPCODE_LOGOUT_RESPONSE'};
		iscsi_pdu_target_sn(\%pdu,1);
		iscsi_pdu_logout_response(\%pdu, \$packet);
		if (!iscsi_net_send($socket, \$packet))
		{
			logger_err("portald_err> unable to send to network?");
			return undef;
		}
		return 1;
	}		

	logger_err("portald_err> expecting Logout Request got opcode: $pdu{'opcode'}");
	return undef;
}




#------------------------------------------------------------------------------
# iportal track pid
#------------------------------------------------------------------------------
sub iportal_daemon_pid
{
	my $pid = $$;
	open (FH, ">$pid_file") or die "Unable to open pid file: $pid_file reason: $!";
	logger_info("portald> launching pid: $pid");
	print FH "$pid";
	close FH;
}
sub iportal_daemon_kill
{
	open (FH, "<$pid_file") or die "Unable to open pid file: $pid_file reason: $!";
	my $pid = <FH>;
	logger_info("portald> killing pid: $pid");
	kill SIGTERM, $pid;
	close FH;
	unlink($pid_file);
}

