#
# iscsi-portal utilities
#
# Steffen Plotner, 2010
#
package iscsiportal;
use strict;
require Exporter;

use Cwd qw(abs_path);
use Data::Dumper;
$Data::Dumper::Indent = 1;
use Sys::Syslog;

our @ISA = qw(Exporter);
our @EXPORT = qw(
	initiator_discover_targets
	initiator_target_portal
);

use config;
use logutil;
use iscsi;
use lockutil;

#
# daemonized processes require absolute filenames
#
my $state_dir = abs_path("$cfg{'state_directory'}");

#------------------------------------------------------------------------------
# given a initiator name and a destination discovery ip address, determine
# all related target portal locations, query them using SendTargets=All with
# given initiator name and return a set of functional targets and their network
# portals.
#------------------------------------------------------------------------------
sub initiator_discover_targets
{
	my ($discovery_ip, $initiator) = @_;

	#
	# given the discovery ip, track the set of accepted portal addresses as a hash
	#
	my %portal_addresses;
	foreach my $portal_address ( @{ $cfg{'discovery_portals'}{$discovery_ip}{'portal_addresses'} } )
	{
		$portal_addresses{ $portal_address } = undef;
	}
	#print Dumper \%portal_addresses;
	#exit;
	
	#
	# now enumerate over all known portals and query them with a SendTargets=All
	# also track a list of portals that did not respond, we need to eliminate dead portals
	#
	my %discovered_targets;
	my @dead_portal_addresses;
	foreach my $portal_address ( @{ $cfg{'discovery_portals'}{$discovery_ip}{'portal_addresses'} } )
	{
		my $portal_group = $cfg{'discovery_portals'}{$discovery_ip}{'portal_group_name'};
		my $portal_port = $cfg{'discovery_portals'}{$discovery_ip}{'portal_port'};
		
		#
		# ask each portal for a list of available targets (yes, you will
		# get some of the answers multiple times if a target is exposed
		# on multiple IP addresses, for example - however, make sure
		# that a targetname is not used twice on different hosts!!!!)
		#
		# iet will only return target addresses that are active at the portal,
		# if even if targets.allow has specified a series of IP addresses for
		# a given target, only those IP addresses that are actually active at the
		# target are handed out (i.e. if one interface is down, we won't get that
		# in the discovered_targets response - and this is only TRUE IF the interface
		# on the target is actually gone and down, if it is up and there is a network
		# issue, then this is FALSE.)
		#
		logger("discover> discovery_ip: $discovery_ip ---> $portal_address:$portal_port");
		
		my $targets = iscsi_initiator_sendtargets_all($initiator, $portal_address, $portal_port);
		#print Dumper \$targets;
		if ( !defined($targets) )
		{
			#
			# track a non-responding portal address
			#
			logger_err("discover_err> dead portal address: $portal_address");
			push @dead_portal_addresses, $portal_address;
		}
		
		my $target_name;
		logger("discover> initiator: $initiator portal: $portal_address:$portal_port");
		foreach my $target ( @{$targets} )
		{
			logger("discover> $target");
			
			if ( $target =~ /^TargetName=(\S+)$/ )
			{
				$target_name = $1;
				$discovered_targets{$target_name};
			}
			if ( defined($target_name) && 
				$target =~ /^TargetAddress=(\S+):(\S+),(\S+)$/ )
			{
				my $discover_portal_address = $1;
				my $discover_portal_port = $2;
				my $discover_tpgt = $3;
				my $discover_portal = "$discover_portal_address:$discover_portal_port,$discover_tpgt";

				#
				# verify that the portal_address and portal_port is valid
				# in the sense, that we know about it in config.pm
				#
				if ( !exists ( $portal_addresses{ $discover_portal_address } ) ||
					$discover_portal_port ne $portal_port )
				{
					logger_err("discover_err> found new unconfigured portal: $discover_portal on target: $target, ignoring it");
				}
				else
				{
					#
					# a single host can expose a target on multiple portal
					# addresses and portal ports.
					#
					$discovered_targets{$target_name}{'portals'}{$discover_portal} =
						{
							"portal_address" => $discover_portal_address,
							"portal_port" => $discover_portal_port,
							"tpgt" => $discover_tpgt,
						};
					$discovered_targets{$target_name}{'portal_group'} = $portal_group;
				}
			}
		}
	}
	#print Dumper \%discovered_targets;
	#exit;

	#  'iqn.1990-01.edu.amherst:iet.iscsi-target3.pg-vlan141-iscsi.vg-satabeast0-vol0.lv-test' => {
	#    'portals' => {
	#      '148.85.141.75:3260,1' => {
	#        'portal_address' => '148.85.141.75',
	#        'tpgt' => '1',
	#        'portal_port' => '3260'
	#      },
	#      '148.85.141.76:3260,1' => {
	#        'portal_address' => '148.85.141.76',
	#        'tpgt' => '1',
	#        'portal_port' => '3260'
	#      },
	#      '148.85.141.74:3260,1' => {
	#        'portal_address' => '148.85.141.74',
	#        'tpgt' => '1',
	#        'portal_port' => '3260'
	#      }
	#    },
	#    'portal_group' => 'pg_vlan141_iscsi'
	#  },

	#
	# eliminate from the discovered_targets hash, portals that are known to be non-reachable
	#
	foreach my $dead_portal_address ( @dead_portal_addresses )
	{
		foreach my $target ( keys %discovered_targets )
		{
			foreach my $portal ( keys %{ $discovered_targets{$target}{'portals'} } )
			{
				my $portal_address = $discovered_targets{$target}{'portals'}{$portal}{'portal_address'};
				if ( $portal_address eq $dead_portal_address )
				{
					logger("discover: removing dead portal $portal");
					delete $discovered_targets{$target}{'portals'}{$portal};
				}
			}
		}
	}
	#print Dumper \%discovered_targets;
	return \%discovered_targets;
}

#------------------------------------------------------------------------------
# given a source ip, destination (discovery ip), initiator and destination
# target, determine an ip address the initiator can use, if none found, return
# nothing.
#------------------------------------------------------------------------------
sub initiator_target_portal
{
	my ($s_ip, $d_ip, $s_initiator, $d_target) = @_;

	#
	# create initiator/target state directory (make this safe - there should be NO slashes
	# in those names(!))
	#
	my $s_initiator_clean = $s_initiator;
	my $d_target_clean = $d_target;
	$s_initiator_clean =~ s/\//_/g;
	$d_target_clean =~ s/\//_/g;
	
	my $i_t_dir = "$state_dir/$s_initiator";
	mkdir($i_t_dir);
	
	my $i_t_resource = "$state_dir/$s_initiator_clean/$d_target_clean";
	my $i_t_data_file = "$i_t_resource.data";
	
	#
	# lock resource (i_t nexus)
	#
	my $lock = resource_lock($i_t_resource);
	if ( !defined($lock) )
	{
		logger_err("portal_err> unable to obtain resource lock $i_t_resource");
		return undef;
	}

	#
	# establish the initiator target portal data
	#
	my %i_t_data;
	my $error = 0;
	eval
	{
		if ( ! -e $i_t_data_file )
		{
			$i_t_data{'target'} = $d_target;
			$i_t_data{'initiator'} = $s_initiator;

			$i_t_data{'portals'} = undef;
			$i_t_data{'last_portal_count'} = 0;
			$i_t_data{'last_portal_index'} = -1;
		}
		else
		{
			i_t_data_read(\%i_t_data, $i_t_data_file);
		}
		i_t_data_elect(\%i_t_data, $d_ip, $s_initiator, $d_target);
		i_t_data_write(\%i_t_data, $i_t_data_file);
	};
	if ( $@ )
	{
		$error = 1;
		logger_err("portal_err> critical section error: $@");
	}
	
	#print Dumper \%i_t_data;

	#$1 = {
	#  'initiator' => 'iqn.1994-05.com.redhat:swiscsi.ietdebug.iscsigrey1.y2zx0s458m03741nv1975jax',
	#  'last_refresh' => 1292644628,
	#  'target' => 'iqn.1990-01.edu.amherst:iet.iscsi-target4.pg-vlan141-iscsi.vg-satabeast5-vol0.lv-ietdebug-sqlgrey1-vol00',
	#  'last_portal_index' => 6,
	#  'portals' => {
	#    '148.85.141.72:3260,1' => {
	#      'portal_address' => '148.85.141.72',
	#      'tpgt' => '1',
	#      'portal_port' => '3260'
	#    },
	#    '148.85.141.73:3260,1' => {
	#      'portal_address' => '148.85.141.73',
	#      'tpgt' => '1',
	#      'portal_port' => '3260'
	#    }
	#  },
	#  'last_portal_count' => 2
	#};
	
	#
	# unlock resource
	#
	resource_unlock($lock, $i_t_resource);
	
	#
	# mod last_portal_index with the last_portal_count, we can return a portal,
	# if the last_portal_count is zero, then we could not find a portal,
	# delay that response.
	#
	my $target_portal;
	if ( $i_t_data{'last_portal_count'} == 0 || $error )
	{
		#
		# delay the error response
		#
		logger("portal> pausing redirect response by 2 seconds");
		select(undef,undef,undef,2);

		logger_access($s_ip, "redirect to 'undef' (REDIR_ERROR)");
		$target_portal = undef;
	}
	else
	{
		#
		# prepare the target_portal response
		#
		my $portal_index = ($i_t_data{'last_portal_index'} % $i_t_data{'last_portal_count'});
		my @target_portal_keys = sort keys %{$i_t_data{'portals'}};
		my $target_portal_key = $target_portal_keys[$portal_index];
		
		my $target_ip =   $i_t_data{'portals'}{$target_portal_key}{'portal_address'};
		my $target_port = $i_t_data{'portals'}{$target_portal_key}{'portal_port'};
		
		logger_access($s_ip, "redirect to $target_ip (REDIR_NORMAL)");
		$target_portal = "$target_ip:$target_port";
	}
	return ($target_portal);
}

#------------------------------------------------------------------------------
# elect next valid portal, or start with a random portal selection
#------------------------------------------------------------------------------
sub i_t_data_elect
{
	my ($i_t_data, $d_ip, $s_initiator, $d_target) = @_;
	
	#
	# refresh either on a last_portal_count equal to zero OR state_time_sec expiration
	#
	if ( $$i_t_data{'last_portal_count'} == 0 || 
		time - $$i_t_data{'last_refresh'} > $cfg{'state_time_sec'})
	{
		#
		# get the targets the initiator can see and store the desired target's 
		# portal addresses (there might be none, during a failover of 
		# resource at the iscsi-target side, for example)
		#
		my $discovered_targets = initiator_discover_targets($d_ip, $s_initiator);

		#
		# must re-init this value, in case there are no available portals for
		# the desired target
		#
		$$i_t_data{'last_portal_count'} = 0;

		if ( exists($$discovered_targets{$d_target}) )
		{
			$$i_t_data{'portals'} = $$discovered_targets{$d_target}{'portals'};
			$$i_t_data{'last_portal_count'} = scalar(keys %{ $$i_t_data{'portals'} });
			$$i_t_data{'last_portal_index'} = int(rand($$i_t_data{'last_portal_count'}));
		}
	
		#
		# set the refresh time to now
		#
		$$i_t_data{'last_refresh'} = time;
	}
	else
	{
		$$i_t_data{'last_portal_index'}++;
	}
}

#------------------------------------------------------------------------------
# write the i_t_data hash to disk
#------------------------------------------------------------------------------
sub i_t_data_write
{
	my ($i_t_data, $i_t_data_file) = @_;

	
	open FOUT, ">$i_t_data_file" or die "Unable to open file: $i_t_data_file";

	$Data::Dumper::Varname = "i_t_data";
	print FOUT Dumper $i_t_data;
	$Data::Dumper::Varname = undef;

	close FOUT;
}

#------------------------------------------------------------------------------
# read the i_t_data hash from disk
#------------------------------------------------------------------------------
sub i_t_data_read
{
	my ($i_t_data, $i_t_data_file) = @_;
			
	open FIN, "<$i_t_data_file" or die "Unable to open file: $i_t_data_file";
			
	my @out = <FIN>;
	my $out = join '', @out;
	close FIN;

	#
	# the data file produces a reference to i_t_data1, that we first initialize
	#
	my $i_t_data1;
	eval $out;

	#
	# copy i_t_data1 hash into i_t_data hash
	#
	foreach my $key (keys %{$i_t_data1} )
	{
		$$i_t_data{$key} = $$i_t_data1{$key};
	}
}
